
# secretary_algorithms.py
# -------------------------------------------------------------
# Stand‑alone implementation of finite‑ and infinite‑horizon
# secretary (best‑choice) problems, featuring:
#   * Backward Induction (finite horizon)
#   * Value Iteration
#   * Policy Iteration
#   * Modified Policy Iteration
# -------------------------------------------------------------
import numpy as np

# ---------- Part A ----------
def finite_secretary(N: int):
    u1 = np.zeros(N + 2)
    u0 = np.zeros(N + 2)
    u1[N] = 1.0
    u0[N] = 0.0
    for t in range(N - 1, 0, -1):
        u0[t] = (1 / (t + 1)) * u1[t + 1] + (t / (t + 1)) * u0[t + 1]
        u1[t] = max(t / N, u0[t])
    r = next(t for t in range(1, N + 1) if t / N >= u0[t])
    p_success = u0[1]
    return r, p_success

# ---------- Helpers ----------
def _solve_threshold(gamma: float, tol: float = 1e-12):
    theta = 0.5
    for _ in range(1000):
        a = (1.0 - theta ** 2) / (2.0 * (1.0 - gamma * theta))
        new_theta = gamma * a
        if abs(new_theta - theta) < tol:
            break
        theta = new_theta
    return theta

# ---------- Part B ----------
def infinite_secretary(gamma: float = 0.99, algo: str = "value", m_eval: int = 5,
                       grid: int = 1001, eps: float = 1e-10):
    xs = (np.arange(grid) + 0.5) / grid
    if algo == "value":
        V = np.zeros(grid)
        while True:
            mean_V = V.mean()
            V_new = np.maximum(xs, gamma * mean_V)
            if np.max(np.abs(V_new - V)) < eps:
                break
            V = V_new
        theta = xs[np.argmax(xs >= gamma * V.mean())]
    elif algo == "policy":
        theta = _solve_threshold(gamma)
    elif algo == "mpi":
        theta = 0.5
        V = xs.copy()
        for _ in range(100):
            for _ in range(m_eval):
                mean_V = V.mean()
                V = np.where(xs >= theta, xs, gamma * mean_V)
            new_theta = gamma * V.mean()
            if abs(new_theta - theta) < 1e-12:
                break
            theta = new_theta
    else:
        raise ValueError("algo must be 'value', 'policy', or 'mpi'")
    p_success = (theta + 1.0) / 2.0
    return theta, p_success

# ---------- Part C ----------
def random_length_secretary(L: float, grid: int = 1001, eps: float = 1e-10):
    xs = (np.arange(grid) + 0.5) / grid
    V = np.zeros(grid)
    while True:
        mean_V = V.mean()
        V_new = np.maximum(xs, L * mean_V)
        if np.max(np.abs(V_new - V)) < eps:
            break
        V = V_new
    theta = xs[np.argmax(xs >= L * V.mean())]
    p_success = (theta + 1.0) / 2.0
    return theta, p_success
